#ifndef HARKIO_ENDIAN_HPP_
#define HARKIO_ENDIAN_HPP_

#include <boost/detail/endian.hpp>
#include <cstddef>
#include <stdint.h>

/** @file endian.hpp
 * @brief endian conversion facilities.
 */

namespace harkio2
{

/** @brief Provides utilities for endian conversion.
 */
class Endian
{
public:
    /** @brief Returns value converted to host's byte order.
     * @param[in] val The value represented in little endian.
     * @return The value represented in host's endian.
     */
    template <typename T>
    static T LittleToHost(T val);

    /** @brief Returns value converted to host's byte order.
     * @param[in] val The value represented in big endian.
     * @return The value represented in host's endian.
     */
    template <typename T>
    static T BigToHost(T val);

    /** @brief Returns value converted to little endian byte order.
     * @param[in] val The value represented in host's endian.
     * @return The value represented in little endian.
     */
    template <typename T>
    static T HostToLittle(T val);

    /** @brief Returns value converted to big endian byte order.
     * @param[in] val The value represented in host's endian.
     * @return The value represented in big endian.
     */
    template <typename T>
    static T HostToBig(T val);

    /** @brief Converts little endian byte order to host's byte order.
     * @param[out] dst Byte array for result.
     * @param[in] src Little endian byte array.
     * @param[in] n The length of byte array.
     */
    static void LittleToHost(void* dst, const void* src, size_t n);

    /** @brief Converts big endian byte order to host's byte order.
     * @param[out] dst Byte array for result.
     * @param[in] src Big endian byte array.
     * @param[in] n The length of byte array.
     */
    static void BigToHost(void* dst, const void* src, size_t n);

    /** @brief Converts host's byte order to little endian byte order.
     * @param[out] dst Byte array for result.
     * @param[in] src Host's endian byte array.
     * @param[in] n The length of byte array.
     */
    static void HostToLittle(void* dst, const void* src, size_t n);

    /** @brief Converts host's byte order to big endian byte order.
     * @param[out] dst Byte array for result.
     * @param[in] src Host's endian byte array.
     * @param[in] n The length of byte array.
     */
    static void HostToBig(void* dst, const void* src, size_t n);

    /** @brief Returns true if this host machine is little endian.
     */
    static bool HostIsLittle()
    {
#if defined(BOOST_LITTLE_ENDIAN)
        return true;
#elif defined(BOOST_BIG_ENDIAN)
        return false;
#elif defined(BOOST_PDP_ENDIAN)
#error PDP endian is not supported
#else
#error BOOST library must define one of BOOST_LITTLE_ENDIAN, BOOST_BIG_ENDIAN, BOOST_PDP_ENDIAN
#endif
    }

    /** @brief Returns true if this host machine is big endian.
     */
    static bool HostIsBig()
    {
#if defined(BOOST_LITTLE_ENDIAN)
        return false;
#elif defined(BOOST_BIG_ENDIAN)
        return true;
#elif defined(BOOST_PDP_ENDIAN)
#error PDP endian is not supported
#else
#error BOOST library must define one of BOOST_LITTLE_ENDIAN, BOOST_BIG_ENDIAN, BOOST_PDP_ENDIAN
#endif
    }

    /** @brief Copies src to dst swapping byte order.
     * Don't overlap each byte array.
     * This function works like memcpy.
     * @param[out] dst A destination byte array.
     * @param[in] src A source byte array.
     * @param[in] n The length of byte array.
     */
    static void MemcpySwap(void* dst, const void* src, size_t n);

};

template <typename T>
T Endian::LittleToHost(T val)
{
    if (HostIsLittle())
    {
        return val;
    }
    else
    {
        T temp;
        LittleToHost(reinterpret_cast<void*>(&temp), reinterpret_cast<void*>(&val), sizeof(T));
        return temp;
    }
}

template <typename T>
T Endian::BigToHost(T val)
{
    if (HostIsBig())
    {
        return val;
    }
    else
    {
        T temp;
        BigToHost(reinterpret_cast<void*>(&temp), reinterpret_cast<void*>(&val), sizeof(T));
        return temp;
    }
}

template <typename T>
T Endian::HostToLittle(T val)
{
    return LittleToHost(val);
}

template <typename T>
T Endian::HostToBig(T val)
{
    return BigToHost(val);
}

}

#endif // HARKIO_ENDIAN_HPP_

